//
//  Threshold_assessment_IBM.cpp
//

/************************************************************************
 Description
 ************************************************************************/
/*
 This is the code for individual-based model designed to study
 the evolution of signaling systems when receivers assess signals
 in a threshold manner and the threshold value of the receivers can evolve.
 */


/************************************************************************
 Includes
 ************************************************************************/

#include <iostream>
#include <vector>
#include <math.h>
#include <random>
#include <algorithm>
#include <chrono>  // for high_resolution_clock

/************************************************************************
 Random Number Generator
 ************************************************************************/
// Psuedo-random number generator.
// It uses a Mersenne Twister algorithm to generate the random numbers.
// Then transforms their distributions accordingly
// You can replace this with any random number generator you want

typedef std::mt19937 RNG; // making typedef for the random number generator

// Uniform distribution
struct Unif {
    double i; // min
    double k; // max
    RNG rng; // random number generator object
    Unif(RNG& rng_) : rng(rng_) {}
    // contructor takes seed
    // can make seed drawn from random device if you want it different every time
    
    inline double doub(double min,double max) { // outputs double
        // takes min and max of the range
        i = min; k = max;
        std::uniform_real_distribution<double> dist(i,k);
        return dist(rng);
    }
    
    inline float flo(double min,double max) { // outputs float
        // takes min and max of the range
        i = min; k = max;
        std::uniform_real_distribution<float> dist(i,k);
        return dist(rng);
    }
    
    inline int in(double min,double max) { // outputs int
        // takes min and max of the range
        i = min; k = max;
        std::uniform_int_distribution<> dist(i,k);
        return dist(rng);
    }
    
    //    void setSeed(int seed){
    //
    //    }
    //
    
};

// Normal Distribution
struct Norm {
    double i; // mean
    double k; // standard deviation
    RNG rng; // random number generator object
    Norm(RNG& rng_) : rng(rng_) {}
    // contructor takes unsigned int seed
    
    inline double doub(double mean,double stdev){ // outputs double
        // takes variables of the mean (i) and stadard deviation (k) for normal distribution
        i = mean; k = stdev;
        std::normal_distribution<double> set(i,k);
        return set(rng);
    }
    
    inline float flo(double mean,double stdev){ //outputs float
        // takes variables of the mean (i) and stadard deviation (k) for normal distribution
        i = mean; k = stdev;
        std::normal_distribution<float> set(i,k);
        return set(rng);
    }
};

/************************************************************************
 Creating random objects
 ************************************************************************/
// creating random object to be used in member functions
// Random numbers are generated from Mersenne Twister algorithm. See RandomWrap.hpp for more info
// Can replace with your own random number generator.

/////// Uniform distribution //////////
RNG mt = std::mt19937(913437950);
// creating random number generator object used to draw random numbers later
// number in parentheses is the seed. It can be changed and also be determined by random device
// RNG is typedef for random number generator defined in RandomWrap.hpp

Unif rUnif(mt); // object for unifrom distribution

/////////  Normal distrubution //////////
RNG mtNorm = std::mt19937(500291813);
// Need different seed for different distibutions so they aren't drawn from same random number list.

Norm rNorm(mtNorm); // object for normal distribution



/************************************************************************
 Creating struct Finch
 ************************************************************************/
// Finches are the agents in the simulation
// Note this model is not specifically designed to fit a finch system, rather finches are just  nice example of a signalling organism

struct Finch {
    ////////  Variables ////////
    double genotype[2][11];
    double tGenotype; // genotype determining the threshold value
    double signal; // males signal value
    bool sex; // true = female, false = male
    bool mated; // true = already mated, they can only mate once per year
    float quality; // quality from 0-1
    float thresh; // threshold value if female
    bool displayed; // true = male has signaled that year
    
    ////////  Functions of Finches ////////
    
    // constructor
    
    Finch(){
        mated = false;
        sex = rUnif.in(0, 1);   // randomly assigning sex of indiviudals
        quality = rUnif.doub(0, 1); // randomly assingly quality to individuals
    }
    
    void setGenotype(float myArr[2][11]){
        // setGenotype - sets the genotype and automatically calculates the phenotype
        // pars - an array of size 10
        
        ////////// Setting signalling genotype /////////
        for (int i=0; i < 2; i++){
            for (int k=0; k < 11; k++){
                genotype[i][k] = myArr[i][k];
            }
        }
        
        
        /////// Calculating signaling value ///////////
        
        int q = floor(10 * quality + .5);
        // evaluating quality with accuracy of .1 and converting to genotype array value
        
        signal = genotype[0][q] + genotype[1][q]+ rNorm.doub(0,.01);
        // first two parts are genetic parts, last part is environmental
        
        if(signal < 0){ // signal cannot be negatitive
            signal = 0;
        }
        
    }
    
    void initate(float mean, float var){
        // initiate - gives initial values for genotype
        // mean - mean value of inital genotype: usually set to threshold / 2
        // divided by 2 because two alleles make the phenotype
        // var - environmental variance in the threshold value, this will equal tVar
        
        // Initial allelic options of each locus
        for (int i=0; i < 11; i++){ // each locus
            for (int j=0; j < 2; j++){ // each allele
                genotype[j][i] = rNorm.doub(mean,.2); // randomly assigned as random value for Norm dist
            }
        }
        
        /////// Calculating signaling value ///////////
        int q = floor(10 * quality + .5);
        // evaluating quality with accuracy of .1 and converting to genotype array value
        
        signal = genotype[0][q] + genotype[1][q]+ rNorm.doub(0,0);
        // first two parts are genetic parts, last part is environmental
        
        if(signal < 0){ // signal cannot be negatitive
            signal = 0;
        }
        
        ////////// Setting threshold genotype and phenotype /////////
        tGenotype = rNorm.doub(mean,.2); // initial threshold value is set at mean
        thresh = tGenotype + rNorm.doub(0, sqrt(var)); // add environmental component to determine phenotype
    }
};


/************************************************************************
 Creating Class Simulation
 ************************************************************************/
// This class is used to run the simulation

class Simulation{
private:
    
    // ****  Parameters *****
    float tBar; // mean threshold value
    float tVar; // variance in threshold value among females
    int K; // maximum number of mating sites
    double u; // per chromosome mutation rate
    double m0; // baseline mortality rate of no signaling
    int B; // offspring produced per mating event
    double alpha; // mutation effect size (standard deviation) manuscript give variance
    int N; // maximum number of signalers evaluated
    
public:
    
    // ****  Population vectors *****
    std::vector <Finch> females; // vector of adult females individuals
    std::vector <Finch> males; // vector of adult males
    std::vector <Finch> maters; // vector of females occupying matings sites
    std::vector <Finch> offspring; // vector that stores offspring before they are added to pop at beginning of next generation
    
    // ****  Functions *****
    void setup(float tBar_, float tVar_, int K_, double u_, double m0_, int B_, double alpha_,int N_){
        // assigns all of the parameter values
        tBar = tBar_;
        tVar = tVar_;
        K = K_;
        u = u_;
        m0 = m0_;
        B = B_;
        alpha = alpha_;
        N = N_;
    }
    
    void init(int size){
        // init - creates the inital population
        // pars - size = the number of individuals in the inital population
        
        females.clear(); // clears previous females
        males.clear(); // clears previous males
        offspring.clear(); // clears previous juveniles
        
        for (int i=0; i < size; i++){
            Finch F;
            
            F.initate((tBar),tVar); // initate individuals with mean phenotype of tBar for all qualities
            
            
            if(F.sex == true){ // if female
                females.push_back(F); // add to females vector
            }
            else{ // if male
                males.push_back(F); // add to males vector
            }
        }
    }
    
    
    void mate(Finch mother,Finch father){
        // mate - creates an offspring
        // pars - mother: female mating, father: male mating
        
        int B2 = B; // number of other produced in this mating event
        
        // with a probability of the father's quality, have one additional offspring
        if(rUnif.doub(0,1) < father.quality) {
            B2 = B2 + 1;
        }
        
        
        for(int i=0; i<B2; i++){
            Finch A;
            
            if(rUnif.doub(0, 1) < 0.5){ // if offspring survives, make offspring
                // Assigning genome
                float g[2][11];
                
                for (int j=0; j < 11; j ++){ // creating mother haplotype
                    g[0][j] = mother.genotype [rUnif.in(0,1)][j];
                }
                
                for (int k=0; k < 11; k++){ // creating father haplotype
                    g[1][k] = father.genotype[rUnif.in(0,1)][k];
                }
                
                /// mutation ///
                if(rUnif.doub(0,1) < u){ // if mutation happens on mother haplotype
                    int alleleM = rUnif.in(0,10);
                    g[0][alleleM] = g[0][alleleM] + rNorm.doub(0,alpha);
                }
                
                if(rUnif.doub(0,1) < u){ // if mutation happens on father haplotype
                    int alleleF = rUnif.in(0,10);
                    g[1][alleleF] = g[1][alleleF] + rNorm.doub(0,alpha);
                }
                
                A.setGenotype(g); // actually setting genotype
                
                // Setting tGenotype (determined by averaging mother and fathers)
                A.tGenotype = ((mother.tGenotype + father.tGenotype) / 2);
                
                // mutation in tGenotype (same mutation rate as signalling phenotype)
                if(rUnif.doub(0,1) < (4*u)){ // if mutation happens (2 time the mutation rate)
                    A.tGenotype = A.tGenotype + rNorm.doub(0,2*alpha);
                }
                
                // Setting threshold
                A.thresh = A.tGenotype + rNorm.doub(0, sqrt(tVar));
                
                // add offspring to the vector of juveniles
                offspring.push_back(A);
            }
        }
    }
    
    
    void step(){
        // step - runs one year of the simulation
        
        //        std::cout << "Males: " << males.size() << std::endl;
        //        std::cout << "Females: " << females.size() << std::endl;
        
        
        //////////////  Mortality //////////////
        int c1 = 0; // counter
        while (c1 < males.size()){
            // probability of mortality determined by signal and quality
            if(rUnif.doub(0, 1) < ((((1 - males[c1].quality)*.5)* (males[c1].signal)) + m0)){
                //                std::cout << males[c1].signal << ", " << males[c1].quality  <<  std::endl;
                males.erase(males.begin() + c1); // erases male c1 from pop
            }
            else{ // if survives move to next ind.
                c1 ++; //
            }
            
        }
        
        int c2 = 0; // counter2
        while (c2 < females.size()){
            if(rUnif.doub(0, 1) < (m0*2)){ // prob of mortality
                females.erase(females.begin() + c2); // erases female c2 from pop
            }
            else{ // if survives move to next ind.
                c2 ++;
            }
        }
        
        
        
        //        std::cout << "Males: " << males.size() << std::endl;
        //        std::cout << "Females: " << females.size() << std::endl;
        
        //////////////  filling of mating sites  //////////////
        // I make the females choose mating sites just be randomly shuffling the order of the
        // female vector. Then, only the first K females get to mate. This is equilivant
        // to randomly sampling without replacement
        
        // randomly rearrange the order of individuals in the population vector
        std::shuffle(females.begin(),females.end(),mt);
        
        
        //////////////  Reproduction  //////////////
        
        int numM = 0; // storing total number of mate males
        
        // selecting mates
        for(int i=0; i < K; i++){ // for each mating female (first K)
            if( (males.size()== 0) || (females.size()== 0)){
                std::cout << "POPULATION WENT EXTINCT" << std::endl;
                break;
            }
            
            int numE = 0; // storing number of unmated males that the female has evaluated
            
            while(numE < N){
                
                int pM = rUnif.in(0,(males.size()-1)); // selecting potential mate
                if(males[pM].mated == false){ // if not mated
                    if(males[pM].signal >= females[i].thresh){  // mate if above threshold
                        mate(females[i], males[pM]);
                        //                        std::cout << "Thresh: " << females[i].thresh << ",  ";
                        //                        std::cout << "Signal: " << males[pM].signal << std::endl;
                        males[pM].mated = true;
                        numM ++;
                        break;
                    } else{   // move to new potential mate
                        numE ++;
                        
                        // but if Nth male, no matter what
                        if(numE == N){
                            mate(females[i], males[pM]);
                        }
                    }
                }
                
            }
            if(numM == males.size()){ // break if all males are mated
                break;
            }
        }
        
        
        ///////////// Update Population ///////////////
        // move offspring to adults
        
        for(int i=0; i < offspring.size(); i++){
            if(offspring[i].sex == true){ // if female add to females
                females.push_back(offspring[i]);
            }
            else{ // if male add to males
                males.push_back(offspring[i]);
            }
        }
        offspring.clear();
        
        // change all males to unmated
        for(int i=0; i < males.size(); i++){
            if(males[i].mated==true){
                males[i].mated = false;
            }
        }
    }
    
    
    
    
    void burnIN(){ // burn in for 10000 years
        // same as step() but mate choice is indepedent of signal
        
        for(int i=0; i < 10000; i++){
            
            
            //////////////  Mortality //////////////
            int c1 = 0; // counter
            while (c1 < males.size()){
                // probability of mortality determined by signal and quality
                if(rUnif.doub(0, 1) < ((((1 - males[c1].quality)*.5)* (males[c1].signal)) + m0)){
                    //                std::cout << males[c1].signal << ", " << males[c1].quality  <<  std::endl;
                    males.erase(males.begin() + c1); // erases male c1 from pop
                }
                else{ // if survives move to next ind.
                    c1 ++; //
                }
                
            }
            
            int c2 = 0; // counter2
            while (c2 < females.size()){
                if(rUnif.doub(0, 1) < (m0*2)){ // prob of mortality
                    females.erase(females.begin() + c2); // erases female c2 from pop
                }
                else{ // if survives move to next ind.
                    c2 ++;
                }
            }
            
            //            std::cout << "Males: " << males.size() << std::endl;
            //            std::cout << "Females: " << females.size() << std::endl;
            
            
            
            
            //////////////  filling of mating sites  //////////////
            // I make the females choose mating sites just be randomly shuffling the order of the
            // female vector. Then, only the first K females get to mate. This is equilivant
            // to randomly sampling without replacement
            
            // randomly rearrange the order of individuals in the population vector
            std::shuffle(females.begin(),females.end(),mt);
            
            
            //////////////  Reproduction  //////////////
            
            int numM = 0; // storing total number of mate male
            
            // mating
            for(int i=0; i < K; i++){ // for each mating female (first K)
                if( (males.size()== 0) || (females.size()== 0)){
                    break;
                }
                
                while(females[i].mated==false){
                    int pM = rUnif.in(0,(males.size()-1)); // selecting potential mate
                    
                    if(males[pM].mated == false){ // if not mated
                        if(rUnif.doub(0, 1) <  .05 + ((4 * males[pM].signal) / (1 + (4*males[pM].signal)))){ // probability of mating
                            
                            mate(females[i], males[pM]);
                            males[pM].mated = true;
                            females[i].mated = true;
                            numM ++;
                        }
                    }
                }
                if(numM == males.size()){ // break if all males are mated
                    break;
                }
            }
            
            
            
            
            ///////////// Update Population ///////////////
            // move offspring to adults
            
            for(int i=0; i < offspring.size(); i++){
                if(offspring[i].sex == true){ // if female add to females
                    females.push_back(offspring[i]);
                }
                else{ // if male add to males
                    males.push_back(offspring[i]);
                }
            }
            offspring.clear();
            
            // change all males to unmated
            for(int i=0; i < males.size(); i++){
                if(males[i].mated==true){
                    males[i].mated = false;
                }
            }
            
            for(int i=0; i < females.size(); i++){
                if(females[i].mated==true){
                    females[i].mated = false;
                }
            }
            //            std::cout << females.size() << std::endl;
        }
        
    }
    
    
    void sim(int gens){ // runs one run of the simulation
        // par gens: number of years
        init(K);
        burnIN();
        for(int i=0; i<gens; i++){
            step();
        }
    }
    
    
    void meanGeno(){ // outputs mean genotype of the population
        double mG[11]; // storing mean genotype
        
        // adding up females
        for(int i=0; i<females.size(); i++){
            for(int j=0; j<11; j++){
                mG[j] += (females[i].genotype[0][j] + females[i].genotype[1][j]);
            }
        }
        
        // adding up males
        for(int i=0; i<males.size(); i++){
            for(int j=0; j<11; j++){
                mG[j] += (males[i].genotype[0][j] + males[i].genotype[1][j]);
            }
        }
        
        // dividing by population size and outputing
        for(int i=0; i<11; i++){
            mG[i] = mG[i] / (females.size() + males.size());
            
            // printing
            std::cout << i << ", " << mG[i] << std::endl;
        }
    }
    
    
    void meanTgeno(){ // outputs mean Tgenotype (threshold genotype) of the population
        double mG = 0; // storing mean genotype
        
        // adding up females
        for(int i=0; i<females.size(); i++){
            mG += females[i].tGenotype;
        }
        
        // adding up males
        for(int i=0; i<males.size(); i++){
            mG += males[i].tGenotype;
        }
        
        // dividing by population size and outputing
        mG = mG / (females.size() + males.size());
        
        // printing
        std::cout << "MeanThreshold, " << mG << std::endl;
        
        
    }
    
    
    void qualSign(){ // outputs quality and signal for each male in the population
        for(int i=0; i<males.size(); i++){
            std::cout << males[i].quality << ", " << males[i].signal << std::endl;
        }
    }
    
    bool extinct(){
        if(males.size()==0){
            return(true);
        }
        if(females.size()==0){
            return(true);
        }
        else{
            return(false);
        }
    }
    
};



/************************************************************************
 Main Function
 ************************************************************************/
int main() {
        float tbar; float tVar; int K; double u; double m0; int B; double alpha; int N; int reps;
    
        std::cout << "tbar, ";
        std::cin >> tbar;
        std::cout << tbar << std::endl;
        std::cout << "tVar, ";
        std::cin >> tVar;
        std::cout << tVar << std::endl;
        std::cout << "K, ";
        std::cin >> K;
        std::cout << K << std::endl;
        std::cout << "u, ";
        std::cin >> u;
        std::cout << u << std::endl;
        std::cout << "m0, ";
        std::cin >> m0;
        std::cout << m0 << std::endl;
        std::cout << "B, ";
        std::cin >> B;
        std::cout << B << std::endl;
        std::cout << "alpha, ";
        std::cin >> alpha;
        std::cout << alpha << std::endl;
        std::cout << "N, ";
        std::cin >> N;
        std::cout << N << std::endl;
        std::cout << "reps, ";
        std::cin >> reps;
        std::cout << reps << std::endl;
    
        // Record start time
        auto start = std::chrono::high_resolution_clock::now();
    
        //    // running simulation
        Simulation test;
        test.setup(tbar, tVar, K, u, m0, B, alpha,N);
    
        std::cout << std::endl;
    
        for(int i=0; i<reps; i++){
            std::cout << "rep" << (i+1) << std::endl;
    
            test.init(K*2); // Initate population
    
            test.burnIN();  // Burn In
    
            for(int j=0; j<20000; j++){  // Running for 20,000 time steps
                test.step();
                if(test.extinct() == true){ // End simulation if population goes extinct
                    break;
                }
            }
    
            test.meanGeno(); // Output the mean signaling genotype of the population
            test.meanTgeno(); // Output the mean threshold genotype of the population
    
            std::cout << std::endl;
            std::cout << std::endl;
        }
    
        // Outputs time the simulation takes to run
        auto finish = std::chrono::high_resolution_clock::now();
        std::chrono::duration<double> elapsed = finish - start;
        std::cout << "Elapsed time: " << elapsed.count() << " s\n";
    
        return 0;

};

